/*
 * HPTalx
 * A PC<->HP48/49 communication program for Linux
 *
 * (C) 1999, 2000 by Bruno Barberi Gnecco
 * (C) 2005 by Bruno Barberi Gnecco, Rafael Ostertag
 *
 * $Id: editor.c,v 1.3 2007/09/29 21:21:01 guengel Exp $
 *
 *
 * Copyright note:
 *
 * This program is free software; you can
 * redistribute it and/or modify it under the
 * terms of the GNU General Public License as
 * published by the Free Software Foundation;
 * either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that
 * it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU
 * General Public License along with this
 * program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 */

#include "hptalx.h"
#include "gui.h"
#include "io.h"
#include "kermit.h"
#include "messages.h"
#include "editor.h"

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <wait.h>

/*
 * Holds the pid of the editor 
 */
static volatile pid_t editor_pid = (-1);

/*
 * The signal set before playing arround in start_editor() 
 */
static sigset_t     old_signal_mask;

/*
 * Saves from where we got the file 
 */
static VPANED       saved_whereami;

/*
 * Directory of the file to be edited 
 */
static char        *editor_directory = NULL;

/*
 * File name of the file to be edited 
 */
static char        *editor_filename = NULL;

/*
 * In case we edit a HP file, stores the path from the HP Calculator 
 */
static char        *editor_remote_directory = NULL;

/*
 * In case we edit a HP file, stores the file name from the HP Calculator 
 */
static char        *editor_remote_filename = NULL;

/*
 * Flag indicating whether or not the editor has already been started 
 */
static volatile int editor_started = 0;

void                signal_handler ();

/*--------------------------------------------------------------------------*/

void
signal_handler ()
{
    pid_t               pid;
    int                 status;

    pid = wait (&status);
    if (pid == editor_pid) {
	hp_file_type        ft;
	callback_data       cb_data;
	int                 cancel_transfer;
	char                filepath[_POSIX_PATH_MAX];

	/*
	 * Reset editor_pid 
	 */
	editor_pid = -1;

	/*
	 * Depending on whether the user edited a file from the HP Calculator 
	 * send it back an clean up 
	 */
	if (saved_whereami != HP) {
	    /*
	     * User edited a file on the PC, so simply clean up allocated
	     * memory 
	     */
	    goto THEEND;
	}

	/*
	 * Compose the file path of the local file 
	 */
	snprintf (filepath, _POSIX_PATH_MAX, "%s/%s", editor_directory,
		  editor_filename);

	/*
	 * Figure out if the editor exited successfully, if not do not
	 * transfer file, only remove it and clean up 
	 */
	if ((WIFEXITED (status)) && (WEXITSTATUS (status) != 0)) {
	    unlink (filepath);
	    goto THEEND;
	}
	/*
	 * Since we are here, we have to transfer the file back to the HP
	 * Calculator and remove the copy on the PC
	 */

	/*
	 * Create the transfer dialog 
	 */
	gui_transfer_dialog_create (&cb_data.dialog, &cb_data.filelabel,
				    &cb_data.infolabel, &cb_data.progressbar,
				    &cancel_transfer);

	/*
	 * Use activity mode for the progress bar 
	 */
	gui_transfer_progress_bar_set_activity (GTK_PROGRESS_BAR
						(cb_data.progressbar));

	/*
	 * The filename of the file to be transfered 
	 */
	gtk_label_set_text (GTK_LABEL (cb_data.filelabel), editor_filename);

	/*
	 * Change to the editor directory on the PC 
	 */
	k_send_command ("LCD %s", editor_directory);
	k_discard_output ();

	/*
	 * Change to the directory on HP Calculator where we got the file
	 * from 
	 */
	hp_chdir (editor_remote_directory);

	/*
	 * Delete the old file 
	 */
	k_send_remote_command ("'%s' PURGE", editor_remote_filename);
	k_discard_output ();

	/*
	 * Get the file type of the file to transfer 
	 */
	hp_get_file_version (filepath, &ft);
	/*
	 * Transfer the file to the hp 
	 */
	hp_send_file (filepath, ft, gui_transfer_callback, &cb_data);

	/*
	 * Remove the file from the PC 
	 */
	unlink (filepath);

	/*
	 * Restore the current pwd on the Calculator 
	 */
	hp_chdir (data_hp.current);

	/*
	 * Restore the current pwd on the PC 
	 */
	k_send_command ("LCD %s", data_pc.current);
	k_discard_output ();

	/*
	 * Destroy the transfer dialog 
	 */
	gtk_widget_destroy (cb_data.dialog);

	/*
	 * Restore the signal process mask 
	 */
	sigprocmask (SIG_SETMASK, &old_signal_mask, NULL);
    } else {
	perror (MSG_ERR_WRONG_PID_SIGCHLD);
	/*
	 * Do nothing, since our editor is still running 
	 */
	return;
    }

  THEEND:
    if (editor_directory != NULL) {
	free (editor_directory);
	editor_directory = NULL;
    }
    if (editor_filename != NULL) {
	free (editor_filename);
	editor_filename = NULL;
    }
    if (editor_remote_directory != NULL) {
	free (editor_remote_directory);
	editor_remote_directory = NULL;
    }
    if (editor_remote_filename != NULL) {
	free (editor_remote_filename);
	editor_remote_filename = NULL;
    }

    editor_started = 0;
}

/*
 * This function takes care of launching the editor. It also downloads a file 
 * from the HP Calculator if the user wishes to edit such a file. 
 */
void
start_editor ()
{
    char                filepath[_POSIX_PATH_MAX];

    if (editor_started == 1) {
	/*
	 * Theres already a editor working 
	 */
	return;
    }

    status_clear ();

    /*
     * Make sure the user don't want to edit a directory on either the PC or
     * the HP 
     */
    if (gui_is_directory_selected ()) {
	status_push (MSG_INF_EDIT_DIRECTORIES);
	goto THEEND;
    }

    saved_whereami = whereami_v;
    if (whereami_v == PC) {
	/*
	 * Check if only one item is selected 
	 */
	if (gui_file_list_single_selection (GTK_TREE_VIEW (files_pc)) != TRUE) {
	    status_push (MSG_WARN_CANNOT_EDIT_MULTIPLE);
	    goto THEEND;
	}
	/*
	 * User wants to edit a file on the PC 
	 */
	editor_directory = strdup (data_pc.current);
	editor_filename = strdup (data_pc.filename);
    } else {
	/*
	 * User wants to edit a file from the HP Calculator 
	 */
	hp_status           hpstat;
	char               *tmp_char;
	callback_data       cb_data;
	int                 cancel_transfer;

	/*
	 * Check if only one item is selected 
	 */
	if (gui_file_list_single_selection (GTK_TREE_VIEW (files_hp)) != TRUE) {
	    status_push (MSG_WARN_CANNOT_EDIT_MULTIPLE);
	    goto THEEND;
	}

	/*
	 * Save the remote directory and file name 
	 */
	editor_remote_directory = strdup (data_hp.current);
	editor_remote_filename = strdup (data_hp.filename);

	/*
	 * Try to get the temporary directory, where later we will download
	 * the file from the HP Calculator for editing into
	 */
	tmp_char = getenv ("TMP");
	if (tmp_char == NULL) {
	    /*
	     * Env Var TMP not defined, lets see what we have
	     */
	    if (access ("/tmp", R_OK | W_OK | X_OK) == 0) {
		editor_directory = strdup ("/tmp");
	    } else if (access ("/var/tmp", R_OK | W_OK | X_OK) == 0) {
		editor_directory = strdup ("/var/tmp");
	    } else {
		editor_directory = strdup (data_pc.current);
	    }
	} else {
	    char               *dir_pos;

	    /*
	     * Get the position of the value in tmp_char 
	     */
	    dir_pos = strstr (tmp_char, "=");
	    if (dir_pos == NULL) {
		/*
		 * What the f*** have we got, fall back to current pwd 
		 */
		editor_directory = strdup (data_pc.current);
	    } else {
		/*
		 * Advance pointer to point right on the directory 
		 */
		dir_pos++;

		/*
		 * Ok, get that path 
		 */
		editor_directory = (char *) malloc (strlen (dir_pos) + 1);
		if (editor_directory == NULL)
		    goto THEEND;
		strncpy (editor_directory, dir_pos, strlen (dir_pos) + 1);
	    }			/* (dir_pos == NULL) */

	}			/* (tmp_char == NULL) */

	/*
	 * Now we got the path where we will download the file into, so get
	 * the file name as well 
	 */
	editor_filename = strdup (data_hp.filename);

	/*
	 * Change the local pwd to the editor_directory 
	 */
	k_send_command ("LCD %s", editor_directory);
	k_discard_output ();

	/*
	 * Prepare the transfer dialog 
	 */
	gui_transfer_dialog_create (&cb_data.dialog, &cb_data.filelabel,
				    &cb_data.infolabel, &cb_data.progressbar,
				    &cancel_transfer);

	/*
	 * Use activity mode for the progress bar 
	 */
	gui_transfer_progress_bar_set_activity (GTK_PROGRESS_BAR
						(cb_data.progressbar));

	/*
	 * Set the label in the transfer dialog to the file the user wants
	 * the edit, hence retrieve 
	 */
	gtk_label_set_text (GTK_LABEL (cb_data.filelabel),
			    editor_remote_filename);

	/*
	 * And finally get the file itself 
	 */
	hpstat =
	    hp_get_file (editor_remote_filename, ttype, gui_transfer_callback,
			 &cb_data);
	if (hpstat != HP_OK) {
	    gtk_widget_destroy (cb_data.dialog);
	    /*
	     * Switch the path back to the current pwd 
	     */
	    k_send_command ("LCD %s", data_pc.current);
	    k_discard_output ();

	    message_box (MSG_ERR,
			 MSG_ERR_CANNOT_RETRIEVE_FILE,
			 editor_remote_filename);

	    goto THEEND;
	}

	/*
	 * Switch the path back to the current pwd 
	 */
	k_send_command ("LCD %s", data_pc.current);
	k_discard_output ();

	/*
	 * Destroy the transfer dialog 
	 */
	gtk_widget_destroy (cb_data.dialog);
    }				/* (whereami == PC) */

    /*
     * Create the filepath for the file to be edited 
     */
    snprintf (filepath, _POSIX_PATH_MAX, "%s/%s", editor_directory,
	      editor_filename);

    /*
     * Fire up the editor 
     */
    editor_pid = fork ();
    if (editor_pid < 0) {
	/*
	 * Remove retrieved file 
	 */
	if (saved_whereami == HP)
	    unlink (filepath);

	message_box (MSG_ERR, MSG_ERR_CANNOT_FORK_EDITOR);
	goto THEEND;
    }

    if (editor_pid == 0) {	/* CHILD */
	execlp (preferences.editor, preferences.editor, filepath, NULL);
	/*
	 * If we come here, there is something wrong 
	 */
	perror (MSG_ERR_CANNOT_EXEC_EDITOR);
	exit (1);
    } else {			/* Parent */
	struct sigaction    sact;
	sigset_t            wdc;
	sigset_t            sset;
	int                 ret_val;

	/*
	 * Setup the signal handler 
	 */
	sigemptyset (&sset);
	sigaddset (&sset, SIGCHLD);
	ret_val = sigprocmask (SIG_UNBLOCK, &sset, &old_signal_mask);
	if (ret_val == (-1)) {
	    message_box (MSG_ERR, MSG_ERR_CANNOT_SET_SIGNAL_MASK);
	}

	/*
	 * Install the signal handler 
	 */
	sigemptyset (&wdc);
	sact.sa_handler = signal_handler;
	sact.sa_mask = wdc;
	sact.sa_flags = SA_NOCLDSTOP | SA_RESTART;

	ret_val = sigaction (SIGCHLD, &sact, NULL);

	editor_started = 1;
    }

    /*
     * We bail out before THEEND, because the signal handler will take care
     * cleaning up things 
     */
    return;

  THEEND:
    if (editor_directory != NULL) {
	free (editor_directory);
	editor_directory = NULL;
    }
    if (editor_filename != NULL) {
	free (editor_filename);
	editor_filename = NULL;
    }
    if (editor_remote_directory != NULL) {
	free (editor_remote_directory);
	editor_remote_directory = NULL;
    }
    if (editor_remote_filename != NULL) {
	free (editor_remote_filename);
	editor_remote_filename = NULL;
    }

    editor_started = 0;

}
