
/*
 * HPTalx
 * A PC<->HP48/49 communication program for Linux
 *
 * (C) 1999, 2000 by Bruno Barberi Gnecco
 * (C) 2005 by Bruno Barberi Gnecco, Rafael Ostertag
 *
 * $Id: io.c,v 1.4 2007/10/06 19:11:07 guengel Exp $
 *
 *
 * Copyright note:
 *
 * This program is free software; you can
 * redistribute it and/or modify it under the
 * terms of the GNU General Public License as
 * published by the Free Software Foundation;
 * either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that
 * it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU
 * General Public License along with this
 * program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "hptalx.h"
#include "gui.h"
#include "kermit.h"
#include "editor.h"
#include "icons.h"
#include "messages.h"
#include "io.h"

#ifdef UPDATE_ROM
#include "hp49flash-0.2.4/upgrade_lib.h"
#endif

#include <string.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>
#include <libgen.h>

/*
 * Flag to needs to be set to 1 in order to make the transfer functions in
 * io.c aware of the abort 
 */
static int          cancel_transfer;

static RETVALUE     _io_check_file_type (const char *f);
static void         _io_populate_pc_files ();
static void         _io_populate_hp_files ();
static void         _io_set_list_pixmap (GtkListStore * list,
										 GtkTreeIter * iter,
										 unsigned short type);

static void         _io_init_my_kermit ();

void                io_command_execute_callback (GtkWidget * w,
												 gpointer data);

/*--------------------------------------------------------------------------*/

/*
 * Checks the file type of the file specified by *f and warns the user about
 * not matching file type/calculator type 
 */
static              RETVALUE
_io_check_file_type (const char *f)
{
	hp_file_type        ft;
	RETVALUE            ret_val;

	hp_get_file_version (f, &ft);

	switch (ft) {
	case HP_FT_UNKNOWN:
		/*
		 * Warn about unknown file type
		 */
		ret_val = confirm (MSG_WARN, MSG_WARN_FILETYPEMISSMATCH, f);
		return ret_val;
		break;
	case HP_FT_ANY:
		/*
		 * Well, that's okay, proceed
		 */
		return OK;
		break;
	default:
		/*
		 * The file appears to be
		 * something we know, but with a
		 * Calculator version attached to
		 * it, so do check.
		 */
		switch (hp_calculator) {
		case HP48:
			if (ft != HP_FT_HP48) {
				ret_val = confirm (MSG_WARN, MSG_WARN_CALCULATORMISSMATCH, f);
				return ret_val;
			}
			return OK;
			break;
		case HP50G:
		case HP49G_PLUS:
		case HP49G:
			if (ft != HP_FT_HP49) {
				ret_val = confirm (MSG_WARN, MSG_WARN_CALCULATORMISSMATCH, f);
				return ret_val;
			}
			return OK;
			break;
		case HP_UNKNOWN:
			/*
			 * To make the compiler stop from complaining about unhandled
			 * enumeration value 
			 */
			return CANCEL;
			break;
		}
	}
	return CANCEL;
}

/*
 * Sets the pixmap according the HP Type. See hpcomm.h #define HP_* and
 * hp_directory_entry struct. 
 */
static void
_io_set_list_pixmap (GtkListStore * list, GtkTreeIter * iter,
					 unsigned short type)
{
	switch (type) {
	case HP_DIRECTORY:

		gtk_list_store_set (list, iter, FLC_TYPE, dir_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_DIRECTORY, -1);
		break;
	case HP_LIST:
		gtk_list_store_set (list, iter, FLC_TYPE, list_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_LIST, -1);
		break;
	case HP_LIBRARY:
		gtk_list_store_set (list, iter, FLC_TYPE, library_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_LIBRARY, -1);
		break;
	case HP_PROGRAM:
		gtk_list_store_set (list, iter, FLC_TYPE, program_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_PROGRAM, -1);
		break;
	case HP_STRING:
		gtk_list_store_set (list, iter, FLC_TYPE, string_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_STRING, -1);
		break;
	case HP_ALG_OBJECT:
		gtk_list_store_set (list, iter, FLC_TYPE, algebraic_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_ALGEBRAIC, -1);
		break;
	case HP_REAL_ARRAY:
		gtk_list_store_set (list, iter, FLC_TYPE, reallist_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_ARRAY, -1);
		break;
	case HP_REAL:
		gtk_list_store_set (list, iter, FLC_TYPE, real_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_REAL, -1);
		break;
	case HP_COMPLEX:
		gtk_list_store_set (list, iter, FLC_TYPE, complex_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_COMPLEX, -1);
		break;
	case HP_COMPLEX_ARRAY:
		gtk_list_store_set (list, iter, FLC_TYPE, complexlist_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_COMPLEX_ARRAY,
							-1);
		break;
	case HP_UNIT_OBJECT:
		gtk_list_store_set (list, iter, FLC_TYPE, unit_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_UNIT, -1);
		break;
	case HP_BINT:
		gtk_list_store_set (list, iter, FLC_TYPE, integer_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_INTEGER, -1);
		break;
	case HP_GLOBAL_NAME:
		gtk_list_store_set (list, iter, FLC_TYPE, global_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_GLOBAL_NAME,
							-1);
		break;
	default:
		gtk_list_store_set (list, iter, FLC_TYPE, unknown_pixbuf, -1);
		gtk_list_store_set (list, iter, FLC_TYPESTR, ROW_DATA_FILE, -1);
		break;
	}
}

/*
 * Populates the files_pc and dir_pc CLists by searching the directory
 * specified in the global struct data_pc using scandir. Depending on the
 * type of the entry (Directory/File) it is either only added to the dir_pc
 * CList (Directories) or files_pc CList (Directories/Files).
 */
static void
_io_populate_pc_files ()
{
	GtkListStore       *list_store_dir;
	GtkListStore       *list_store_file;
	GtkTreeViewColumn  *dir_name_col;
	GtkTreeIter         iter_dir;
	GtkTreeIter         iter_file;
	register int        i = 0;
	struct dirent     **ep;
	struct stat         st;
	int                 n;
	char                size_text[20];
	char                filenamewithpath[BUFFERSIZE];

	status_push (MSG_INF_REFRESH_PC);

	/*
	 * Set the title of the column of the PC Directory List 
	 */
	dir_name_col =
		gtk_tree_view_get_column (GTK_TREE_VIEW (dir_pc), DLC_NAME);
	gtk_tree_view_column_set_title (dir_name_col,
									(const gchar *) data_pc.current);

	n = scandir (data_pc.current, &ep, 0, alphasort);
	if (n < 0) {
		message_box (MSG_ERR, MSG_ERR_CANNOT_SCANDIR);
		return;
	}

	/*
	 * Get the store for the PC Directory List 
	 */
	list_store_dir =
		GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (dir_pc)));
	gtk_list_store_clear (list_store_dir);

	/*
	 * Get the store for the PC File List 
	 */
	list_store_file =
		GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (files_pc)));
	gtk_list_store_clear (list_store_file);

	for (i = 0; i < n; i++) {
		sprintf (filenamewithpath, "%s/%s", data_pc.current, ep[i]->d_name);
		if (stat (filenamewithpath, &st) != 0) {
			status_push (MSG_WARN_CANNOT_STAT_FILE, filenamewithpath);
			continue;
		}
		/*
		 * Acquire that new iterator 
		 */
		gtk_list_store_append (list_store_file, &iter_file);

		/*
		 * Add the filename to the PC File List 
		 */
		gtk_list_store_set (list_store_file, &iter_file, FLC_NAME,
							ep[i]->d_name, -1);

		/*
		 * Add the file size to the PC File List 
		 */
		snprintf (size_text, 20, "%d", (int) st.st_size);
		gtk_list_store_set (list_store_file, &iter_file, FLC_SIZE, size_text,
							-1);

		if (S_ISDIR (st.st_mode)) {
			/*
			 * Add the directory to the PC Directory List 
			 */
			gtk_list_store_append (list_store_dir, &iter_dir);
			gtk_list_store_set (list_store_dir, &iter_dir, DLC_NAME,
								ep[i]->d_name, -1);

			/*
			 * Add the appropriate icon to the PC File List 
			 */
			gtk_list_store_set (list_store_file, &iter_file, FLC_TYPE,
								dir_pixbuf, -1);

			/*
			 * Add the "row data" (yes, coming from the CLIST) to the PC File 
			 * List 
			 */
			gtk_list_store_set (list_store_file, &iter_file, FLC_TYPESTR,
								ROW_DATA_DIRECTORY, -1);

		} else {
			/*
			 * Add the appropriate icon to the PC File List 
			 */
			gtk_list_store_set (list_store_file, &iter_file, FLC_TYPE,
								file_pixbuf, -1);

			/*
			 * Add the "row data" (yes, coming from the CLIST) to the PC File 
			 * List 
			 */
			gtk_list_store_set (list_store_file, &iter_file, FLC_TYPESTR,
								ROW_DATA_FILE, -1);
		}

	}

	free (ep);
}

/*
 * Populates the files_hp and dir_hp CLists using the GList obtained by
 * calling hp_get_directory() passing the directory specified in the global
 * struct data_hp.current. 
 */
void
_io_populate_hp_files ()
{
	GtkListStore       *list_store_dir;
	GtkListStore       *list_store_file;
	GtkTreeViewColumn  *dir_name_col;
	GtkTreeIter         iter_dir;
	GtkTreeIter         iter_file;
	hp_status           hpstat;
	GList              *gl = NULL;
	GList              *gl_save = NULL;
	hp_directory_entry *hpd;
	int                 i = 0;
	char                size_text[20];

	status_push (MSG_INF_REFRESH_HP);

	/*
	 * Set the title of the column of the PC Directory List 
	 */
	dir_name_col =
		gtk_tree_view_get_column (GTK_TREE_VIEW (dir_hp), DLC_NAME);
	gtk_tree_view_column_set_title (dir_name_col,
									(const gchar *) data_hp.current);

	/*
	 * Retrieve the directory entries for the directory in
	 * data_hp.current 
	 */
	hpstat = hp_get_directory (data_hp.current, &gl);
	if (hpstat != HP_OK) {
		gui_print_nice_hpstatus (hpstat);
		goto THEEND;
	}

	/*
	 * Get the store for the PC Directory List 
	 */
	list_store_dir =
		GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (dir_hp)));
	gtk_list_store_clear (list_store_dir);

	/*
	 * Get the store for the PC File List 
	 */
	list_store_file =
		GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (files_hp)));
	gtk_list_store_clear (list_store_file);

	/*
	 * Check if we not retrieved the entries for the { HOME } directory.
	 * If so, add ".." to the dir_hp CList, so the user is able to change 
	 * one directory up.  
	 */
	if (!strstr (data_hp.current, "HOME }")
		&& !strstr (data_hp.current, "HOME}")) {
		gtk_list_store_append (list_store_dir, &iter_dir);
		gtk_list_store_set (list_store_dir, &iter_dir, DLC_NAME, "..", -1);
	}

	gl_save = gl;
	gl = g_list_first (gl);
	i = 0;
	while (gl != NULL) {
		/*
		 * Retrieve the data from the GList 
		 */
		hpd = (hp_directory_entry *) gl->data;

		/*
		 * Acquire that new iterator 
		 */
		gtk_list_store_append (list_store_file, &iter_file);

		/*
		 * Add the filename to the PC File List 
		 */
		gtk_list_store_set (list_store_file, &iter_file, FLC_NAME, hpd->name,
							-1);

		/*
		 * Add the file size to the PC File List 
		 */
		snprintf (size_text, 20, "%li", hpd->size);
		gtk_list_store_set (list_store_file, &iter_file, FLC_SIZE, size_text,
							-1);

		/*
		 * This sets the pixbuf as well as the "row data" 
		 */
		_io_set_list_pixmap (list_store_file, &iter_file, hpd->type);

		if (hpd->type == HP_DIRECTORY) {
			gtk_list_store_append (list_store_dir, &iter_dir);
			gtk_list_store_set (list_store_dir, &iter_dir, DLC_NAME,
								hpd->name, -1);

		}

		i++;
		free (hpd->name);
		free (hpd);
		gl = g_list_next (gl);

	}

  THEEND:
	if (gl_save != NULL)
		g_list_free (gl_save);

}

/*
 * The master function responsible for transfering data to and from the HP
 * Calculator
 */
void
io_transfer (GtkWidget * w, gpointer dorefresh)
{
	k_status            kstat;
	hp_status           hpstat;
	GList              *selection_list;
	GList              *selection_list_backup;
	GtkTreeSelection   *tree_selection;
	GtkTreeView        *tree_view;
	GtkTreeModel       *tree_model;
	GtkTreeIter         list_iter;
	gboolean            got_iter;
	callback_data       cb_data;
	char               *t;

	if (connected == OFFLINE)
		return;

	status_clear ();

	/*
	 * Make sure the user don't want to transfer a unix directory to the 
	 * HP Calculator 
	 */
	if ((whereami_v == PC) && (gui_is_directory_selected ())) {
		status_push (MSG_ERR_CANNOT_XFER_DIRS);
		return;
	}

	/*
	 * Make sure the user has not selected an item in one of the directory
	 * lists 
	 */
	if (whereami_h == DIRS) {
		message_box (MSG_INF, MSG_INF_SELECT_IN_FILE_XFER);
		return;
	}

	kstat = k_send_remote_command ("3 TRANSIO");
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		return;
	}
	kstat = k_discard_output ();
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		return;
	}

	status_push (MSG_INF_XFER);

	/*
	 * Initialize the cancel flag 
	 */
	cancel_transfer = 0;

	gui_transfer_dialog_create (&cb_data.dialog, &cb_data.filelabel,
								&cb_data.infolabel, &cb_data.progressbar,
								&cancel_transfer);

	/*
	 * Change current working directory
	 */
	kstat = k_send_command ("LCD %s", data_pc.current);
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		gtk_widget_destroy (cb_data.dialog);
		return;
	}

	kstat = k_discard_output ();
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		gtk_widget_destroy (cb_data.dialog);
		return;
	}

	/*
	 * Get the right tree view 
	 */
	if (whereami_v == PC) {
		/*
		 * Once again, we cannot transfer directories to the HP Calculator,
		 * a previous checks took place to make sure the user doesn't want
		 * to transfer a unix directory to the HP Calculator, so there is no
		 * need to check again, therefore assume tree_view is files_pc
		 */
		tree_view = GTK_TREE_VIEW (files_pc);
	} else {
		tree_view = GTK_TREE_VIEW (files_hp);
	}

	/*
	 * Get the tree selection 
	 */
	tree_selection = gtk_tree_view_get_selection (tree_view);
	selection_list =
		gtk_tree_selection_get_selected_rows (tree_selection, &tree_model);
	if ((selection_list == NULL) || (selection_list->data == NULL)) {
		message_box (MSG_ERR, MSG_ERR_NO_ITEMS_SELECTED);
		gtk_widget_destroy (cb_data.dialog);
		return;
	}

	selection_list_backup = selection_list = g_list_first (selection_list);
	while (selection_list != NULL) {
		char               *file_size_text;
		unsigned long       file_size;
		char                filename[_POSIX_PATH_MAX];

		/*
		 * Acquire the iteratore for the list 
		 */
		got_iter =
			gtk_tree_model_get_iter (tree_model, &list_iter,
									 (GtkTreePath *) selection_list->data);
		if (got_iter != TRUE) {
			message_box (MSG_ERR, MSG_ERR_NO_ITER_ACQUIRED);
			g_list_foreach (selection_list_backup, (GFunc) gtk_tree_path_free,
							NULL);
			g_list_free (selection_list_backup);
			gtk_widget_destroy (cb_data.dialog);
			return;
		}

		/*
		 * Get the name of the item 
		 */
		t = gui_file_list_get_name_by_iter (tree_model, &list_iter);
		if ((strcmp (t, ".") == 0) || strcmp (t, "..") == 0)
			goto THENEXTPLEASE;

		/*
		 * Here, there is no need to check again if we are in the files or
		 * directory panel on the PC, because that has been taken care of
		 * outside this loop. Transfering HP Directories to the PC works, so
		 * this check is sufficient. 
		 */
		if (whereami_v == PC) {
			hp_file_type        ft;

			/*
			 * This is needed, because if the user selected a region, a
			 * directory might sneak in 
			 */
			if (!strcmp
				(ROW_DATA_DIRECTORY,
				 (char *) gui_file_list_get_type_str_by_iter (tree_model,
															  &list_iter))) {
				status_push (MSG_ERR_CANNOT_XFER_DIRS);
				goto THENEXTPLEASE;
			}

			/*
			 * Check whether the file we want to
			 * transfer to the calculator matches
			 * the calculator version, if not,
			 * inform the user and ask what to
			 * do.
			 */
			snprintf (filename, _POSIX_PATH_MAX, "%s/%s", data_pc.current, t);
			if (_io_check_file_type (filename) != OK)
				goto THENEXTPLEASE;

			/*
			 * Prepare the Transfer dialog 
			 */
			gtk_label_set_text (GTK_LABEL (cb_data.filelabel), t);

			/*
			 * Get the file size, so we can prepare the progress bar on the
			 * Transfer Dialog 
			 */
			file_size_text =
				gui_file_list_get_size_str_by_iter (tree_model, &list_iter);
			file_size = (unsigned long) atol (file_size_text);

			/*
			 * Inform the status bar about the values 
			 */
			gtk_progress_configure (GTK_PROGRESS (cb_data.progressbar),
									0, 0, file_size);

			/*
			 * Check if the user aborted the transfer 
			 */
			if (cancel_transfer == 1)
				goto TRANSFERCANCELLED;

			/*
			 * Since we are here, the file to
			 * transfer matched the calculator,
			 * or the user in all his wisdom
			 * decided to transfer it anyway.
			 */
			hpstat = hp_get_file_version (filename, &ft);
			if (hpstat != HP_OK) {
				gui_print_nice_hpstatus (hpstat);
				goto THENEXTPLEASE;
			}
			hpstat =
				hp_send_file (filename, ft, gui_transfer_callback, &cb_data);
			if (hpstat != HP_OK) {
				gui_print_nice_hpstatus (hpstat);
				goto THENEXTPLEASE;
			}
		} else {
			/*
			 * Check if the user aborted the transfer 
			 */
			if (cancel_transfer == 1)
				goto TRANSFERCANCELLED;

			if (access (data_pc.current, W_OK) != 0) {
				message_box (MSG_ERR, MSG_ERR_NO_W_PERMISSION,
							 data_pc.current);
				goto TRANSFERCANCELLED;
			}

			/*
			 * Prepare the Transfer dialog 
			 */
			gtk_label_set_text (GTK_LABEL (cb_data.filelabel), t);

			/*
			 * Get the file size, so we can prepare the progress bar on the
			 * Transfer Dialog 
			 */
			file_size_text =
				gui_file_list_get_size_str_by_iter (tree_model, &list_iter);
			file_size = (unsigned long) atol (file_size_text);

			/*
			 * Inform the status bar about the values 
			 */
			gtk_progress_configure (GTK_PROGRESS (cb_data.progressbar),
									0, 0, file_size);
			hpstat = hp_get_file (t, ttype, gui_transfer_callback, &cb_data);
			if (hpstat != HP_OK) {
				gui_print_nice_hpstatus (hpstat);
				goto THENEXTPLEASE;
			}

		}

		/*
		 * Check if the user aborted the transfer 
		 */
		if (cancel_transfer == 1)
			goto TRANSFERCANCELLED;

	  THENEXTPLEASE:
		selection_list = g_list_next (selection_list);
	}

	g_list_foreach (selection_list_backup, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (selection_list_backup);

	gtk_widget_destroy (cb_data.dialog);
	if (GPOINTER_TO_INT (dorefresh) == TRUE)
		io_refresh (w, GINT_TO_POINTER ((whereami_v == PC ? HP : PC)));
	return;

  TRANSFERCANCELLED:
	status_push (MSG_INF_CANCEL);
	gtk_widget_destroy (cb_data.dialog);
	g_list_foreach (selection_list_backup, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (selection_list_backup);
	if (GPOINTER_TO_INT (dorefresh) == TRUE)
		io_refresh (w, GINT_TO_POINTER ((whereami_v == PC ? HP : PC)));
	return;
}

/*
 * Function for moving files on the PC or the HP Calculator using the
 * existing functions transfer() and delete() 
 */
void
io_move (GtkWidget * w, gpointer wdc)
{

	if (connected == OFFLINE)
		return;

	status_clear ();

	if (gui_is_directory_selected ()
		&& (!((whereami_v == HP) && (whereami_h == FILES)))) {
		status_push (MSG_INF_MOVE_DIRECTORIES);
		return;
	}

	status_push (MSG_INF_MOVING);
	io_transfer (w, FALSE /* Do not refresh after transfer */ );
	io_delete (w, FALSE /* Do not refresh after deletion */ );
	status_push (MSG_INF_DONE);
	io_refresh (w, GINT_TO_POINTER (BOTH));
}

/*
 * Allows the user to edit a file by forking a new process 
 */
void
io_edit (GtkWidget * w, gpointer wdc /* we don't care */ )
{
	/*
	 * Editing file is handled by start_editor() in editor.c 
	 */
	start_editor ();
}

/*
 * Copy files on the PC or the HP Calculator 
 */
void
io_copy (GtkWidget * w, gpointer wdc /* we don't care */ )
{
	GList              *selection_list;
	GList              *selection_list_backup;
	GtkTreeSelection   *tree_selection;
	GtkTreeView        *tree_view;
	GtkTreeModel       *tree_model;
	GtkTreeIter         list_iter;
	gboolean            got_iter;

	hp_status           hpstat;
	RETVALUE            answer;
	struct stat         src_stat;
	char               *t;
	char                src_file[_POSIX_PATH_MAX];
	char                dest_file[_POSIX_PATH_MAX];
	int                 fd_src;
	int                 fd_dest;
	int                 ret_val;
	ssize_t             b_read;
	ssize_t             b_wrote;
	void               *copy_buff;
	char               *newname;

	status_clear ();

	/*
	 * Make sure the user doesn't want to copy a directory on the PC 
	 */
	if ((whereami_v == PC) && (gui_is_directory_selected ())) {
		status_push (MSG_INF_COPY_DIRECTORIES);
		return;
	}

	/*
	 * Make sure the user is not in one of the directory clists 
	 */
	if (whereami_h == DIRS) {
		status_push (MSG_INF_SELECT_IN_FILE_COPY);
		return;
	}

	/*
	 * Ask for the destination directory 
	 */
	newname = entry_popup (MSG_ENTRY, MSG_ENTRY_COPY_DEST_DIR);
	if (!newname)
		return;

	/*
	 * Get the right tree 
	 */
	tree_view =
		(whereami_v ==
		 PC ? GTK_TREE_VIEW (files_pc) : GTK_TREE_VIEW (files_hp));
	/*
	 * Get the tree selection 
	 */
	tree_selection = gtk_tree_view_get_selection (tree_view);
	selection_list =
		gtk_tree_selection_get_selected_rows (tree_selection, &tree_model);
	if ((selection_list == NULL) || (selection_list->data == NULL)) {
		message_box (MSG_ERR, MSG_ERR_NO_ITEMS_SELECTED);
		return;
	}

	selection_list_backup = selection_list = g_list_first (selection_list);
	while (selection_list != NULL) {
		/*
		 * Acquire the iteratore for the list 
		 */
		got_iter =
			gtk_tree_model_get_iter (tree_model, &list_iter,
									 (GtkTreePath *) selection_list->data);
		if (got_iter != TRUE) {
			message_box (MSG_ERR, MSG_ERR_NO_ITER_ACQUIRED);
			g_list_foreach (selection_list_backup, (GFunc) gtk_tree_path_free,
							NULL);
			g_list_free (selection_list_backup);
			return;
		}

		/*
		 * Get the name of the item 
		 */
		t = gui_file_list_get_name_by_iter (tree_model, &list_iter);
		if ((strcmp (t, ".") == 0) || strcmp (t, "..") == 0)
			goto THENEXTPLEASE;

		/*
		 * Here, there is no need to check again if we are in the files or
		 * directory panel on the PC, because that has been taken care of
		 * outside this loop. Transfering HP Directories to the PC works, so
		 * this check is sufficient. 
		 */
		if (whereami_v == PC) {
			/*
			 * This is needed, because if the user selected a region, a
			 * directory might sneak in 
			 */
			if (!strcmp
				(ROW_DATA_DIRECTORY,
				 gui_file_list_get_type_str_by_iter (tree_model,
													 &list_iter))) {
				status_push (MSG_INF_COPY_DIRECTORIES);
				goto THENEXTPLEASE;
			}

			/*
			 * Compose source file path
			 */
			snprintf (src_file, _POSIX_PATH_MAX, "%s/%s", data_pc.current, t);
			/*
			 * Open source file
			 */
			fd_src = open (src_file, O_RDONLY);
			if (fd_src < 0) {
				message_box ("Error", "Could not open file %s.", src_file);
				return;
			}

			/*
			 * Get stats of source file, used for
			 * permissions on new file
			 */
			ret_val = fstat (fd_src, &src_stat);
			if (ret_val < 0) {
				message_box (MSG_ERR, MSG_WARN_CANNOT_STAT_FILE, src_file);
				close (fd_src);
				return;
			}

			/*
			 * Compose destination file path
			 */
			snprintf (dest_file, _POSIX_PATH_MAX, "%s/%s", newname, t);

			/*
			 * Create destination file
			 */
			fd_dest =
				open (dest_file, O_WRONLY | O_CREAT | O_EXCL,
					  src_stat.st_mode);
		  REDO:
			if (fd_dest < 0) {
				switch (errno) {
				case EACCES:
					message_box (MSG_ERR, MSG_ERR_CANNOT_CREATE, dest_file);
					close (fd_src);
					return;
					break;
				case ENOENT:
					message_box (MSG_ERR, MSG_ERR_DEST_DIR_NONEXIST);
					close (fd_src);
					return;
					break;
				}

				answer = confirm (MSG_CONF, MSG_CONF_OVWR, dest_file);
				if (answer == OK) {
					fd_dest = open (dest_file, O_WRONLY | O_TRUNC);
					/*
					 * Ugly hack, but should do it
					 * for now
					 */
					goto REDO;
				} else {
					close (fd_src);
					close (fd_dest);
					return;
				}
			}

			/*
			 * Allocate copy buffer
			 */
			copy_buff = malloc (COPYBUFFERSIZE);
			if (copy_buff == NULL) {
				close (fd_src);
				close (fd_dest);
				/*
				 * In this case, this probably wont
				 * work
				 */
				message_box (MSG_FAT, MSG_FAT_MALLOC);
				return;
			}

			status_push (MSG_INF_COPY_PROGRESS, t, newname);
			/*
			 * Here it comes, copying data over
			 */
			while ((b_read = read (fd_src, copy_buff, COPYBUFFERSIZE)) > 0) {
				b_wrote = write (fd_dest, copy_buff, b_read);
				if (b_wrote != b_read) {
					message_box (MSG_ERR, MSG_ERR_COPY, src_file, dest_file);
					close (fd_src);
					close (fd_dest);
					free (copy_buff);
					return;
				}
			}

			free (copy_buff);
			close (fd_src);
			close (fd_dest);
		} else {
			status_push (MSG_INF_COPY_PROGRESS, t, newname);
			hpstat = hp_copy_file (t, newname);
			if (hpstat != HP_OK)
				gui_print_nice_hpstatus (hpstat);
		}
	  THENEXTPLEASE:
		selection_list = g_list_next (selection_list);
	}

	io_refresh (w, GINT_TO_POINTER (whereami_v));
}

/*
 * Rename a file on the PC or the HP Calculator 
 */
void
io_renamew (GtkWidget * w, gpointer wdc /* we don't care */ )
{
	hp_status           hpstat;
	char               *newname;
	int                 ret_val;

	status_clear ();

	/*
	 * Make sure the user doesn't want to rename a directory on the PC 
	 */
	if ((whereami_v == PC) && (gui_is_directory_selected ())) {
		status_push (MSG_ERR_CANNOT_REN_DIRS);
		return;
	}

	/*
	 * Ask for the new name 
	 */
	newname = entry_popup (MSG_ENTRY, MSG_ENTRY_RENAME);

	if (!newname)
		return;

	if (whereami_v == PC) {
		char               *oldname;

		if (whereami_h == FILES) {
			if (!data_pc.filename)
				return;
		} else {
			if (!data_pc.dir)
				return;
		}
		if ((whereami_h == FILES)
			&& (gui_file_list_single_selection (GTK_TREE_VIEW (files_pc)) !=
				TRUE)) {
			status_push (MSG_WARN_CANNOT_RENAME_MULTIPLE);
			return;
		}
		oldname = (whereami_h == FILES ? data_pc.filename : data_pc.dir);
		ret_val = rename (oldname, newname);
		if (ret_val == -1) {
			message_box (MSG_ERR, MSG_ERR_RENAME, oldname, newname);
		}

	} else {

		/*
		 * Find out which HP panel we are 
		 */
		if (whereami_h == FILES) {
			if (!data_hp.filename)
				return;

			if (gui_file_list_single_selection (GTK_TREE_VIEW (files_hp)) !=
				TRUE) {
				status_push (MSG_WARN_CANNOT_RENAME_MULTIPLE);
				return;
			}
			/*
			 * Figure out whether the user wants to rename a directory or 
			 * variable 
			 */
			if (strstr (data_hp.file_type_str, ROW_DATA_DIRECTORY) != NULL) {
				/*
				 * User wants to rename a directory 
				 */
				hpstat = hp_rename (data_hp.filename, newname, HP_REN_DIR);
				if (hpstat != HP_OK)
					gui_print_nice_hpstatus (hpstat);
			} else {

				/*
				 * User wants to rename a variable 
				 */
				hpstat = hp_rename (data_hp.filename, newname, HP_REN_VAR);
				if (hpstat != HP_OK)
					gui_print_nice_hpstatus (hpstat);
			}
		} else {
			if (!data_hp.dir)
				return;

			hpstat = hp_rename (data_hp.dir, newname, HP_REN_DIR);
			if (hpstat != HP_OK)
				gui_print_nice_hpstatus (hpstat);
		}
	}
	io_refresh (w, GINT_TO_POINTER (whereami_v));
	free (newname);
}

/*
 * Unpacks a zip, tar.gz or tar.bz2 archive on the PC 
 */
void
io_unpack (GtkWidget * w, gpointer wdc /* we don't care */ )
{
	char                temp[BUFFERSIZE];

	if (!data_pc.filename || whereami_v != PC)
		return;

	if (strstr (data_pc.filename, ".zip")
		|| strstr (data_pc.filename, ".ZIP")) {
		sprintf (temp, "cd %s; unzip %s", data_pc.current, data_pc.filename);
		system (temp);
	} else if (strstr (data_pc.filename, ".tgz") ||
			   strstr (data_pc.filename, ".TGZ") ||
			   strstr (data_pc.filename, ".tar.gz") ||
			   strstr (data_pc.filename, ".TAR.GZ")) {
		sprintf (temp, "cd %s; tar zxvf %s", data_pc.current,
				 data_pc.filename);
		system (temp);
	} else if (strstr (data_pc.filename, ".tar.bz2") ||
			   strstr (data_pc.filename, ".TAR.BZ2")) {
		sprintf (temp, "cd %s; tar yxvf %s", data_pc.current,
				 data_pc.filename);
		system (temp);
	} else {
		status_push (MSG_WARN_ARCH_UNKNOWN);
		return;
	}
	status_push (MSG_INF_UNPACK_OK);
	io_refresh (w, GINT_TO_POINTER (PC));
}

/*
 * Deletes a file on the PC or the HP Calculator 
 */
void
io_delete (GtkWidget * w, gpointer dorefresh)
{
	hp_status           hpstat;
	GList              *selection_list;
	GList              *selection_list_backup;
	GtkTreeSelection   *tree_selection;
	GtkTreeView        *tree_view;
	GtkTreeModel       *tree_model;
	GtkTreeIter         list_iter;
	gboolean            got_iter;
	char                filepath[_POSIX_PATH_MAX];
	char               *row_data;
	char               *t;
	int                 remove;
	int                 ret_val;

	status_clear ();

	tree_view =
		(whereami_v ==
		 PC ? GTK_TREE_VIEW (files_pc) : GTK_TREE_VIEW (files_hp));
	/*
	 * Get the tree selection 
	 */
	tree_selection = gtk_tree_view_get_selection (tree_view);
	selection_list =
		gtk_tree_selection_get_selected_rows (tree_selection, &tree_model);
	if ((selection_list == NULL) || (selection_list->data == NULL)) {
		message_box (MSG_ERR, MSG_ERR_NO_ITEMS_SELECTED);
		return;
	}

	selection_list_backup = selection_list = g_list_first (selection_list);
	while (selection_list != NULL) {
		/*
		 * Acquire the iteratore for the list 
		 */
		got_iter =
			gtk_tree_model_get_iter (tree_model, &list_iter,
									 (GtkTreePath *) selection_list->data);
		if (got_iter != TRUE) {
			message_box (MSG_ERR, MSG_ERR_NO_ITER_ACQUIRED);
			g_list_foreach (selection_list_backup, (GFunc) gtk_tree_path_free,
							NULL);
			g_list_free (selection_list_backup);
			return;
		}
		/*
		 * Get the name of the item 
		 */
		t = gui_file_list_get_name_by_iter (tree_model, &list_iter);

		/*
		 * Before we do anything, check if the
		 * user selected a directory in the file
		 * list.
		 */
		row_data =
			gui_file_list_get_type_str_by_iter (tree_model, &list_iter);
		if (strcmp (row_data, ROW_DATA_DIRECTORY) == 0) {
			/*
			 * Olà, dealing with directory, but not in this function.
			 */
			status_push ("Not deleting directory %s", t);
			goto THENEXTPLEASE;
		}

		/*
		 * On pc, construct full filepath 
		 */
		if (whereami_v == PC) {
			snprintf (filepath, _POSIX_PATH_MAX, "%s/%s", data_pc.current, t);
		} else {
			/*
			 * Obviously on hp calc 
			 */
			snprintf (filepath, _POSIX_PATH_MAX, "%s", t);
		}

		if (preferences.delete_confirm == TRUE) {
			remove = confirm (MSG_CONF, MSG_CONF_DEL, filepath);
			if (remove == CANCEL)
				goto THENEXTPLEASE;
		}

		if (whereami_v == PC) {
			ret_val = unlink (filepath);
			if (ret_val == -1) {
				message_box (MSG_ERR, MSG_ERR_DELETE, filepath);
			}

		} else {
			hpstat = hp_purge (t, HP_PURGE_VAR);
			if (hpstat != HP_OK) {
				gui_print_nice_hpstatus (hpstat);
				goto THENEXTPLEASE;
			}
		}

	  THENEXTPLEASE:
		selection_list = g_list_next (selection_list);
	}
	g_list_foreach (selection_list_backup, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (selection_list_backup);

	if (GPOINTER_TO_INT (dorefresh) == TRUE)
		io_refresh (w, GINT_TO_POINTER (whereami_v));
}

/*
 * Pops up a dialog where the user can enter the directory where he wishes to 
 * change to 
 */
void
io_choose_dir (GtkTreeViewColumn * wdc, gpointer refresh_what)
{
	hp_status           hpstat;
	char               *newdir;

	/*
	 * Get the directory name from the user where to change to 
	 */
	newdir = entry_popup (MSG_ENTRY, MSG_ENTRY_CHOOSE_DIR);

	if (!newdir)
		return;

	status_clear ();

	if ((VPANED) GPOINTER_TO_INT (refresh_what) == PC) {
		if (chdir (newdir) != 0) {
			message_box (MSG_ERR, MSG_ERR_DIR_NOT_FOUND, newdir);
			return;
		}
		free (data_pc.current);
		data_pc.current = getcwd (NULL, 0);
	} else {
		if (connected == OFFLINE)
			return;
		hpstat = hp_chdir (newdir);
		if (hpstat != HP_OK)
			gui_print_nice_hpstatus (hpstat);

		free (data_hp.current);
		data_pc.current = strdup (newdir);

	}
	io_refresh (NULL, refresh_what);
}

/*
 * Changes to a new directory on the PC or the HP Calculator 
 */
void
io_change_dir (GtkWidget * wdc1, gpointer wdc2 /* we don't care */ )
{
	hp_status           hpstat;
	char                newdir[BUFFERSIZE];

	if (!gui_is_directory_selected ())
		return;

	status_clear ();

	if (whereami_v == PC) {
		/*
		 * Check if only one item is selected 
		 */
		if (gui_file_list_single_selection (GTK_TREE_VIEW (files_pc)) != TRUE) {
			status_push (MSG_WARN_CANNOT_CHANGE_DIR_MULTIPLE);
			return;
		}
		chdir (whereami_h == FILES ? data_pc.filename : data_pc.dir);
		free (data_pc.current);
		data_pc.current = getcwd (NULL, 0);
	} else {
		/*
		 * Check if only one item is selected 
		 */
		if (gui_file_list_single_selection (GTK_TREE_VIEW (files_pc)) != TRUE) {
			status_push (MSG_WARN_CANNOT_CHANGE_DIR_MULTIPLE);
			return;
		}
		/*
		 * Does the user want to go up one directory and hence clicked
		 * dot-dot in the HP Directory Clist? However, we have to make sure
		 * the user is in the directory clist, because he could have made a
		 * click to dot-dot and then a click to a different directory in the
		 * files clist.  
		 */
		if (strstr (data_hp.dir, "..") && whereami_h == DIRS) {
			hpstat = hp_directory_pop (data_hp.current, newdir, BUFFERSIZE);
			if (hpstat != HP_OK)
				gui_print_nice_hpstatus (hpstat);
			free (data_hp.current);
			data_hp.current = strdup (newdir);

		} else {
			hpstat =
				hp_directory_push (data_hp.current,
								   (whereami_h ==
									FILES ? data_hp.filename : data_hp.dir),
								   newdir, BUFFERSIZE);
			if (hpstat != HP_OK)
				gui_print_nice_hpstatus (hpstat);
			free (data_hp.current);
			data_hp.current = strdup (newdir);
		}
	}
	io_refresh (NULL, GINT_TO_POINTER (whereami_v));
}

/*
 * Creates a directory on the PC or the HP Calculator 
 */
void
io_create_dir (GtkWidget * w, gpointer wdc /* we don't care */ )
{
	hp_status           hpstat;
	char               *newdir;

	/*
	 * Get the name of the new directory 
	 */
	newdir = entry_popup (MSG_ENTRY, MSG_ENTRY_MKDIR);
	if (!newdir)
		return;

	status_clear ();

	if (whereami_v == PC) {
		/*
		 * On PC 
		 */
		if (mkdir (newdir, S_IRWXU | S_IRGRP | S_IWGRP | S_IROTH) != 0)
			message_box (MSG_ERR, MSG_ERR_CANNOT_MKDIR, newdir);
	} else {
		/*
		 * On HP Calculator 
		 */
		hpstat = hp_mkdir (newdir);
		if (hpstat != HP_OK)
			gui_print_nice_hpstatus (hpstat);
	}
	io_refresh (w, GINT_TO_POINTER (whereami_v));
	free (newdir);
}

/*
 * Removes a directory on the PC or the HP Calculator 
 */
void
io_remove_dir (GtkWidget * w, gpointer dorefresh)
{
	hp_status           hpstat;

	if (!gui_is_directory_selected ())
		return;

	status_clear ();

	if (preferences.delete_confirm != FALSE) {
		int                 remove;

		remove = confirm (MSG_CONF, MSG_CONF_DEL_DIR);
		if (remove == CANCEL)
			return;
	}

	if (whereami_v == PC) {
		char                command[512];
		int                 remove;

		snprintf (command, 512, "rm -rf %s/%s", data_pc.current,
				  (whereami_h == FILES ? data_pc.filename : data_pc.dir));

		remove = confirm (MSG_CONF, MSG_CONF_CONF_DEL_DIR, command);
		if (remove != OK)
			return;
		system (command);
	} else {
		/*
		 * On HP Calculator 
		 */
		hpstat =
			hp_purge ((whereami_h == FILES ? data_hp.filename : data_hp.dir),
					  HP_PURGE_DIR);
		if (hpstat != HP_OK)
			gui_print_nice_hpstatus (hpstat);
	}

	if (GPOINTER_TO_INT (dorefresh) == TRUE)
		io_refresh (w, GINT_TO_POINTER (whereami_v));
}

/*
 * This is the function its address is passed to gui_command_window_create() in
 * command(). It is connected to the Send button on the Command Window. data
 * is a pointer to a struct command_window_send_args. 
 */
void
io_command_execute_callback (GtkWidget * w, gpointer argument)
{
	hp_status           hpstat;
	GtkTextBuffer      *buffer;
	char               *output_text;

	hpstat = hp_command (CWSA_P (argument)->command, &output_text);
	if (hpstat != HP_OK)
		gui_print_nice_hpstatus (hpstat);

	buffer =
		gtk_text_view_get_buffer (GTK_TEXT_VIEW (CWSA_P (argument)->output));
	gtk_text_buffer_set_text (buffer, (gchar *) output_text, -1);
	free (output_text);
}

/*
 * Executes the command entered on the HP Calculator 
 */
void
io_command (GtkWidget * w, gpointer wdc /* we don't care */ )
{

	if (connected == OFFLINE)
		return;

	gui_command_window_create (G_CALLBACK (io_command_execute_callback));

}

#ifdef UPDATE_ROM

/*
 * code of rom_upload is from Upgrade, by Mathias 
 * Bunte, and is in the hp49flash-(VERSION)
 * directory. 
 */
void
update_rom (GtkWidget * w, gpointer wdc /* we don't care */ )
{
	char                temp[BUFFERSIZE];
	int                 update;

	if (data_pc.filename == NULL) {
		message_box (MSG_ERR, MSG_ERR_NO_ROM_SELECTED);
		return;
	}
	update = confirm (MSG_CONF, MSG_CONF_ROM_UPDATE);

	if (update != OK)
		return;
	sprintf (temp, "%s/%s", data_pc.current, data_pc.filename);
	rom_upload (temp, preferences.port);
}
#endif

/*
 * Makes a backup of the entire RAM of the HP Calculator 
 */
void
io_backup (GtkWidget * w, gpointer wdc /* we don't care */ )
{
	hp_status           hpstat;
	k_status            kstat;
	GtkWidget          *fs_dialog = NULL;
	RETVALUE            dialog_ret;
	gchar              *backupname = NULL;
	const gchar        *tmp_backupname = NULL;
	char               *backupdir;
	char               *backupfile;
	callback_data       cb_data;

	if (connected == OFFLINE)
		return;

	status_clear ();

	/*
	 * Initialize the dialog member, because in case of an error, a goto to
	 * ERROR_THEEND could be issued, but it might be that at that time the
	 * Transfer Dialog has not been created so far. 
	 */
	cb_data.dialog = NULL;

	/*
	 * Setup the file selection dialog 
	 */

	fs_dialog = gui_backup_dialog_create (&dialog_ret);

	if (dialog_ret != OK) {
		gtk_widget_destroy (fs_dialog);
		status_push (MSG_INF_BACKUP_ABORTED);
		return;
	}

	tmp_backupname =
		gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs_dialog));
	if (!tmp_backupname) {
		gtk_widget_destroy (fs_dialog);
		return;
	}

	backupname = strdup (tmp_backupname);
	/*
	 * Get file name and directory name 
	 */
	backupfile = basename (backupname);
	backupdir = dirname (backupname);

	status_push (MSG_INF_BACKING_UP);

	/*
	 * Change to directory selected by the user on the PC and make our struct
	 */
	free (data_pc.current);
	data_pc.current = strdup (backupdir);
	kstat = k_send_command ("LCD %s", data_pc.current);
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}
	kstat = k_discard_output ();
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}

	kstat = k_send_remote_command ("{ -40 } CF");
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}
	kstat = k_discard_output ();
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}

	/*
	 * Store the HP Calculator flags 
	 */
	status_push (MSG_INF_STORE_FLAGS);

	hpstat = hp_store_flags ();
	if (hpstat != HP_OK) {
		gui_print_nice_hpstatus (hpstat);
		/*
		 * Try to remove the stored flags 
		 */
		hp_purge_stored_flags ();
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}

	/*
	 * Put the name of the backup on the stack of the HP Calculator 
	 */
	kstat = k_send_remote_command (":IO:%s", backupfile);
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		/*
		 * Try to remove the stored flags 
		 */
		hp_purge_stored_flags ();
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}
	kstat = k_discard_output ();
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		/*
		 * Try to remove the stored flags 
		 */
		hp_purge_stored_flags ();
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}

	/*
	 * get the closest to ARCHIVE we can 
	 */
	kstat = k_send_remote_command ("\\\\<< ARCHIVE SERVER");
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		/*
		 * Try to remove the stored flags 
		 */
		hp_purge_stored_flags ();
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}
	kstat = k_discard_output ();
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		/*
		 * Try to remove the stored flags 
		 */
		hp_purge_stored_flags ();
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}

	/*
	 * Now we have to disconnect the HP Calculator and put Kermit into Server 
	 * Mode 
	 */
	kstat = k_send_command ("FINISH");
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		/*
		 * Try to remove the stored flags 
		 */
		hp_purge_stored_flags ();
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}

	kstat = k_discard_output ();
	if (kstat != K_OK) {
		/*
		 * We do not try try remove the stored flags, because the calculator
		 * probably has shutdown its server 
		 */
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}

	/*
	 * run the Kermit server 
	 */
	kstat = k_send_command ("SERVER");
	if (kstat != K_OK) {
		/*
		 * We do not try try remove the stored flags, because the calculator
		 * probably has shutdown its server 
		 */
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}

	/*
	 * Prompt the user for pressing <EVAL> on the Caclulator, so it starts to 
	 * send data over to the PC 
	 */
	message_box ("User interaction", MSG_MSG_BACKUP_START);
	status_push (MSG_INF_BACKING_UP);

	/*
	 * Initialize the cancel flag 
	 */
	cancel_transfer = 0;

	/*
	 * Create the transfer Dialog 
	 */
	gui_transfer_dialog_create (&cb_data.dialog, &cb_data.filelabel,
								&cb_data.infolabel, &cb_data.progressbar,
								&cancel_transfer);

	gui_transfer_progress_bar_set_activity (GTK_PROGRESS_BAR
											(cb_data.progressbar));

	/*
	 * Set the file label to the name of the backup file 
	 */
	gtk_label_set_text (GTK_LABEL (cb_data.filelabel), backupfile);

	/*
	 * Read the progress of the file transfer 
	 */
	kstat = k_read_server_transfer (gui_transfer_callback, &cb_data);
	if (kstat != K_OK) {
		/*
		 * We do not try try remove the stored flags, because the calculator
		 * has shutdown its server 
		 */
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}

	/*
	 * This is because if we send k_interrupt() immediately after
	 * k_read_server_transfer() returned, the Calculator waits for some data, 
	 * which never will arrive, since the Kermit server has been killed, so
	 * we wait for letting them finish their chit-chat.
	 *
	 * Better solutions are welcome.
	 */
	sleep (1);

	/*
	 * send CTRL+C to get out of Kermit server mode
	 */
	kstat = k_interrupt ();
	if (kstat != K_OK) {
		/*
		 * We do not try try remove the stored flags, because the calculator
		 * has shutdown its server 
		 */
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_BACKUP_ABORTED);
		goto THEEND;
	}

	/*
	 * Remove the stored HP Calculator flags 
	 */
	status_push (MSG_INF_REMOVE_STORED_FLAGS);
	hpstat = hp_purge_stored_flags ();
	if (hpstat != HP_OK)
		gui_print_nice_hpstatus (hpstat);

  THEEND:
	/*
	 * And finally destroy the dialogs
	 */
	if (fs_dialog != NULL)
		gtk_widget_destroy (fs_dialog);
	if (cb_data.dialog != NULL)
		gtk_widget_destroy (cb_data.dialog);
	if (backupname != NULL)
		free (backupname);

	io_refresh (w, GINT_TO_POINTER (PC));
}

/*
 * Restores the RAM of the HP Calculator according a backup file created by
 * backup() 
 */
void
io_restore (GtkWidget * w, gpointer wdc /* we don't care */ )
{
	hp_status           hpstat;
	k_status            kstat;
	GtkWidget          *fs_dialog;
	RETVALUE            dialog_ret;
	callback_data       cb_data;
	hp_file_type        ft;
	gchar              *backupname = NULL;
	const gchar        *tmp_backupname = NULL;
	char               *backupdir = NULL;
	char               *backupfile = NULL;
	char               *filename = NULL;

	/*
	 * Warn the user about the implications of that action 
	 */
	dialog_ret = confirm (MSG_CONF, MSG_CONF_RESTORE);
	if (dialog_ret != OK)
		return;

	status_clear ();

	fs_dialog = gui_backup_dialog_create (&dialog_ret);
	if (dialog_ret != OK) {
		gtk_widget_destroy (fs_dialog);
		status_push (MSG_INF_RESTORE_ABORTED);
		return;
	}

	tmp_backupname =
		gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs_dialog));
	if (!tmp_backupname) {
		gtk_widget_destroy (fs_dialog);
		return;
	}

	/*
	 * Make a copy of backupname 
	 */
	filename = strdup (tmp_backupname);

	/*
	 * Get file name and directory name
	 */
	backupname = strdup (tmp_backupname);
	backupfile = basename (backupname);
	backupdir = dirname (backupname);

	/*
	 * Initialize the cancel flag 
	 */
	cancel_transfer = 0;

	/*
	 * Create the File Transfer Dialog 
	 */
	gui_transfer_dialog_create (&cb_data.dialog, &cb_data.filelabel,
								&cb_data.infolabel, &cb_data.progressbar,
								&cancel_transfer);

	/*
	 * Use activity mode for the progress bar 
	 */
	gui_transfer_progress_bar_set_activity (GTK_PROGRESS_BAR
											(cb_data.progressbar));

	/*
	 * Set the file label to the name of the backup file 
	 */
	gtk_label_set_text (GTK_LABEL (cb_data.filelabel), backupfile);

	/*
	 * Change the current working directory 
	 */
	free (data_pc.current);
	data_pc.current = strdup (backupdir);
	kstat = k_send_command ("LCD %s", backupdir);
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_RESTORE_ABORTED);
		goto THEEND;
	}
	kstat = k_discard_output ();
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_RESTORE_ABORTED);
		goto THEEND;
	}

	status_push (MSG_INF_SENDING_BACKUP);
	/*
	 * Check the file type, however _io_check_file_type will pop up a message
	 * box if the file does not match, so we don't have to care about that
	 * and simply can go to THEEND 
	 */
	if (_io_check_file_type (filename) != OK)
		goto THEEND;

	/*
	 * Eventually send the file to the Calculator 
	 */
	hpstat = hp_get_file_version (filename, &ft);
	if (hpstat != HP_OK) {
		gui_print_nice_hpstatus (hpstat);
		message_box (MSG_ERR, MSG_ERR_RESTORE_ABORTED);
		goto THEEND;
	}
	hpstat = hp_send_file (backupfile, ft, gui_transfer_callback, &cb_data);
	if (hpstat != HP_OK) {
		gui_print_nice_hpstatus (hpstat);
		message_box (MSG_ERR, MSG_ERR_RESTORE_ABORTED);
		goto THEEND;
	}

	/*
	 * Check if the user aborted the transfer 
	 */
	if (cancel_transfer == 1)
		goto THEEND;

	status_push (MSG_INF_RESTORING);

	kstat = k_send_remote_command ("'%s' RCL", backupfile);
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_RESTORE_ABORTED);
		goto THEEND;
	}

	kstat = k_discard_output ();
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_RESTORE_ABORTED);
		goto THEEND;
	}

	kstat = k_send_remote_command ("RESTORE");
	if (kstat != K_OK) {
		gui_print_nice_kstatus (kstat);
		message_box (MSG_ERR, MSG_ERR_RESTORE_ABORTED);
		goto THEEND;
	}

	/*
	 * Make sure the command makes it to the Calculator before we cut the
	 * connection 
	 */
	sleep (2);

	io_disconnect (NULL, K_FAST_QUIT);
	message_box (MSG_INF, MSG_INF_AFTER_RESTORING);

  THEEND:

	/*
	 * Finally destroy the Dialogs 
	 */
	if (fs_dialog != NULL)
		gtk_widget_destroy (fs_dialog);
	if (cb_data.dialog != NULL)
		gtk_widget_destroy (cb_data.dialog);
	if (filename != NULL)
		free (filename);
	if (backupname != NULL)
		free (backupname);

	io_refresh (w, GINT_TO_POINTER (PC));
}

void
io_refreshall (GtkWidget * w, gpointer wdc /* we don't care */ )
{
	if (connected == ONLINE)
		io_refresh (w, GINT_TO_POINTER (BOTH));
	else
		io_refresh (w, GINT_TO_POINTER (PC));
}

void
io_refresh (GtkWidget * wdc, gpointer what)
{

	if ((VPANED) GPOINTER_TO_INT (what) == PC ||
		(VPANED) GPOINTER_TO_INT (what) == BOTH) {
		_io_populate_pc_files ();
	}
	if (((VPANED) GPOINTER_TO_INT (what) == HP ||
		 (VPANED) GPOINTER_TO_INT (what) == BOTH) && connected == ONLINE) {
		_io_populate_hp_files ();
	}

	status_clear ();
}

static void
_io_init_my_kermit ()
{
	k_send_command ("SET MODEM TYPE NONE");
	k_discard_output ();

	/*
	 * Those are the settings from the 
	 * Kermit roboust macro
	 */
	k_send_command ("SET REC PACK 90");
	k_discard_output ();
	k_send_command ("SET PREFIXING ALL");
	k_discard_output ();
	k_send_command ("SET RELIABLE OFF");
	k_discard_output ();
	k_send_command ("SET CLEAR-CHANNEL OFF");
	k_discard_output ();

	k_send_command ("SET CARRIER-WATCH OFF");
	k_discard_output ();

	k_send_command ("SET FLOW NONE");
	k_discard_output ();

	k_send_command ("SET PARITY NONE");
	k_discard_output ();

	k_send_command ("SET BLOCK-CHECK 3");
	k_discard_output ();

	k_send_command ("SET CONTROL PREFIX ALL");
	k_discard_output ();

	k_send_command ("SET FILE TYPE BINARY");
	k_discard_output ();

	k_send_command ("SET FILE COLLISION OVERWRITE");
	k_discard_output ();

	k_send_command ("SET FILE NAMES LITERAL");
	k_discard_output ();
}

/*
 * Connects the HP Calculator 
 */
void
io_connect (GtkWidget * w, gpointer wdc /* we don't care */ )
{
	k_params            kparams;
	k_status            k_ret;
	hp_status           hp_ret;
	char                portname[30];

	if (connected == ONLINE)
		return;

	connected = ONLINE;

	status_clear ();

	status_push (MSG_INF_CONNECTING);

	if (preferences.port > 99) {
		snprintf (portname, 30, "/dev/ttyUSB%d", preferences.port - 100);
	} else {
		snprintf (portname, 30, "/dev/ttyS%d", preferences.port);
	}

	kparams.port = portname;
	kparams.speed = preferences.speed;
	kparams.timeout = preferences.wait_connect;

	k_ret = k_connect (&kparams);
	switch (k_ret) {
	case K_OK:
		break;
	case K_CONNECTED:
		break;
	case K_TIMEOUT:
		status_push (MSG_WARN_TIMEOUT);
		message_box (MSG_ERR, MSG_ERR_TIMEOUT_CONNECTING);
		connected = OFFLINE;
		return;
		break;
	default:
		gui_print_nice_kstatus (k_ret);
		connected = OFFLINE;
		return;
		break;
	}

	/*
	 * Doing initialization stuff neccessary for the HP Calculator 
	 */
	_io_init_my_kermit ();

	status_push (MSG_INF_GET_CALC_TYPE);
	hp_ret = hp_version (&hp_calculator);
	if (hp_ret != HP_OK) {
		status_push (MSG_INF_UNABLE_TO_CONNECT);
		message_box (MSG_ERR, MSG_ERR_UNABLE_TO_CONNECT);
		connected = OFFLINE;
		k_quit (K_FAST_QUIT);
		return;
	}
	switch (hp_calculator) {
	case HP48:
		status_push ("HP 48G found");
		break;
	case HP49G_PLUS:
	case HP49G:
		status_push ("HP 49 found");
		break;
	case HP50G:
		status_push ("HP 50 found");
		break;
	case HP_UNKNOWN:
		message_box (MSG_ERR, MSG_ERR_UNKOWN_CALC);
		status_push (MSG_INF_UNKNOWN_CALC);
		connected = OFFLINE;
		k_quit (K_FAST_QUIT);
		return;
	}

	data_hp.filename = data_hp.dir = data_hp.current = NULL;
	data_hp.dir = strdup (HP_HOME);
	data_hp.current = strdup (HP_HOME);

	gui_set_menu_connected ();

	io_refresh (w, GINT_TO_POINTER (HP));
}

/*
 * Disconnects the HP Calculator
 * *how says how to do it, either a fast quit or a full quit
 */
void
io_disconnect (GtkWidget * w, gpointer how)
{
	GtkListStore       *list_store;
	GtkTreeViewColumn  *dir_name_col;

	if (connected == OFFLINE)
		return;

	connected = OFFLINE;
	hp_calculator = HP_UNKNOWN;

	status_clear ();

	status_push (MSG_INF_DISCONNECTING);

	/*
	 * Shutdown kermit 
	 */
	k_quit ((k_quit_mode) how);

	list_store =
		GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (dir_hp)));
	gtk_list_store_clear (list_store);
	list_store =
		GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (files_hp)));
	gtk_list_store_clear (list_store);

	dir_name_col =
		gtk_tree_view_get_column (GTK_TREE_VIEW (dir_hp), DLC_NAME);
	gtk_tree_view_column_set_title (dir_name_col, "Not Connected");

	gui_set_menu_disconnected ();

	if (data_hp.filename)
		free (data_hp.filename);
	if (data_hp.dir)
		free (data_hp.dir);
	if (data_hp.current)
		free (data_hp.current);
	status_push (MSG_INF_DISCONNECTED);
}
