
/*
 *  HPTalx
 *  A PC<->HP48/49 communication program for Linux
 *
 *  (C) 1999, 2000 by Bruno Barberi Gnecco
 *  (C) 2005 by Bruno Barberi Gnecco, Rafael Ostertag
 *
 *
 * $Id: kermit.h,v 1.3 2007/09/29 21:21:01 guengel Exp $
 *
 *
 * Copyright note:
 *
 * This program is free software; you can
 * redistribute it and/or modify it under the
 * terms of the GNU General Public License as
 * published by the Free Software Foundation;
 * either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that
 * it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU
 * General Public License along with this
 * program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 */

#ifndef KERMIT_H
#define KERMIT_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include<stdlib.h>

#define K_KERMIT_PROMPT "Kermit>"	/* The Kermit Prompt */
#define K_TRANSFER_OK "[OK]"	/* The string indicating the transfer has
								 * been successful */

#define K_CHAR_NEWLINE '\n'		/* The newline character */
#define K_STR_NEWLINE "\n"		/* The newline character as string */

/*
 * Kermit Commands 
 */
#define K_CMD_REMOTE_HOST "REMOTE HOST "	/* The Kermit command "REMOTE
											 * HOST " (including the
											 * trailling space) */

#define K_CMD_REMOTE_DIRECTORY  "REMOTE DIRECTORY"	/* The Kermit command for 
													 * getting ls the remote
													 * directory */

#define K_CMD_SET_PROMPT "SET PROMPT"	/* Sets the default prompt */

#define K_CMD_QUIET_ON "SET QUIET ON"	/* Quiet mode */

#define K_CMD_HINTS_OFF "SET HINTS OFF"	/* Kermit hints off */

#define K_CMD_LINE "SET LINE %s"	/* Sets the line used to communicate with 
									 * the device */

#define K_CMD_SPEED "SET SPEED %d"	/* Sets the port speed used for the line */

#define K_CMD_FILE_DISPLAY_CRT "SET FILE DISPLAY CRT"	/* No fancy output */

#define K_CMD_FINISH "FINISH"	/* Finishes server on remote end */

#define K_CMD_CLOSE "CLOSE"		/* Closes connection */

#define K_CMD_QUIT "QUIT"		/* Exits Kermit */

#define K_CMD_SEND_SINGLE_FILE_TEXT "SEND /TEXT %s"	/* Sends a single file in 
													 * text mode */

#define K_CMD_SEND_SINGLE_FILE_BINARY "SEND /BINARY %s"	/* Sends a single
														 * file in binary
														 * mode */

/*
 * Sizes above 512 makes read() and write() to fail
 */
#define K_BUFFERSIZE 512

/*
 * Is used for bulk data reading
 */
#define K_CHUNKSIZE 32

/*
 * The return values of kermit functions
 */
enum _k_status
{
	K_OK = 0,					/* Status OK */
	K_CONNECTED,				/* Indicates that Kermit is connected to the
								 * HP Calculator */
	K_NOTCONNECTED,				/* Indicating that Kermit is not connected to 
								 * the HP Calculator */
	K_FORKPTY,					/* Error forking a pty */
	K_EXECKERMIT,				/* Error exec'ing Kermit */
	K_FCNTL,					/* Error while calling fcntl() */
	K_PIPE,						/* Broken pipe */
	K_INITKERM,					/* Error initializing Kermit */
	K_MALLOC,					/* malloc() error, insufficient memory */
	K_INVARG,					/* Invalid arguments passed to k_* function */
	K_WRITE,					/* write() error */
	K_READ,						/* read() error */
	K_EOF,						/* End of File encountered */
	K_TIMEOUT,					/* Timeout occurred */
	K_SELECT,					/* select() error */
	K_STREAM,					/* stream error (fopen(), fgets(), etc.) */
	K_CANCEL,					/* Transfer has been canceled */
	K_BUFFER,					/* Buffer too small to hold string */
	K_UNKNOWN					/* Unknown error */
};
typedef enum _k_status k_status;

enum _k_transfer_type
{
	AUTO = 0, BINARY, ASCII
};
typedef enum _k_transfer_type k_transfer_type;

enum _k_quit_mode
{
	/*
	 * This will not send a "FINISH" to the
	 * remote host
	 */
	K_FAST_QUIT,
	/*
	 * Sends a "FINISH" to the remote host
	 */
	K_FULL_QUIT
};
typedef enum _k_quit_mode k_quit_mode;

enum _k_cancel_transfer_switches
{
	K_XFER_PROCEED,
	K_XFER_CANCEL
};
typedef enum _k_cancel_transfer_switches k_cancel_transfer_switches;

/*
 * The parameters passed to k_connect() 
 */
struct _k_params
{
	char               *port;	/* Name of the port to be used */
	unsigned int        speed;	/* Port speed */
	int                 timeout;	/* Timeout waiting for response */
};
typedef struct _k_params k_params;

typedef void        (*k_transfer_callback) (unsigned int *bytes,
											unsigned int *p,
											unsigned int *cps, void *data);

/*
 * The global variable holding the last encountered status of Kermit 
 */
extern k_status     k_errno;

/*
 * The global variable indicating whether the current transfer should be
 * canceled 
 */
extern k_cancel_transfer_switches k_cancel_transfer;

/*
 * Holds the transfer mode switches. Used to translate the k_transfer_type
 * values to strings.
 * 
 * Value                        Meaning
 * ------------------------------------------------------------------
 * k_transfer_file_switches[0]  Automatic transfer mode.
 * k_transfer_file_switches[1]  Binary mode. 
 * k_transfer_file_switches[2]  Text mode. 
 */
extern const char  *k_transfer_file_switches[];

extern k_status     k_connect (const k_params * params);

extern k_status     k_is_connected ();

extern k_status     k_quit (k_quit_mode m);

extern k_status     k_discard_output (void);

extern k_status     k_send_command (const char *fmt, ...);

extern k_status     k_send_remote_command (const char *fmt, ...);

extern k_status     k_wait_ready (char **o, size_t * size);

extern k_status     k_read_file_transfer (k_transfer_callback callback,
										  void *data);

extern k_status     k_read_server_transfer (k_transfer_callback callback,
											void *data);

extern k_status     k_free_bulk_output (char **o);

extern k_status     k_interrupt ();

extern k_status     k_cancel_action ();

#define K_CHECKSTATE {if(k_is_connected()==K_NOTCONNECTED) return K_NOTCONNECTED;}

#endif
